use std::{
	path::PathBuf,
	time::SystemTime
};

use actix_web::{
	get, post, web,
	HttpResponse, Responder, HttpRequest, HttpMessage
};

use crate::{
	lib::{
		boards::Board,
		users::{UserCookie, User}
	},
	api::types::*,
	ThreadData
};

use handlebars::to_json;
use maplit::btreemap;
use super::TimeDerive;
use chrono::{DateTime, Utc};
use ammonia::clean;
use pulldown_cmark::{Parser, Options, html::push_html};

#[get("/boards/")]
pub async fn list_boards(
	req: HttpRequest,
	data: web::Data<AppState>,
	tdata: web::Data<ThreadData<'_>>
) -> impl Responder {
	let cookie = req.cookie("auth");
	let user = if let Some(c) = cookie {
		if let Ok(v) = UserCookie::parse(c.value()) {
			data.user_manager.lock().unwrap().get_user_by_id(v.id)
		} else {
			None
		}
	} else {
		None
	};
	let boards = (*data.boards.lock().unwrap()).clone();
	let mut boards_json = to_json(&boards);
	let now = SystemTime::now().duration_since(SystemTime::UNIX_EPOCH).unwrap();
	let boards_json = boards_json.as_array_mut().unwrap().iter_mut().zip(boards).map(|x| -> &serde_json::Value {
		x.0.as_object_mut().unwrap().insert(
			"time_meta".to_string(),
			if let Some(t) = x.1.last_active {
				to_json(std::time::Duration::from_secs(t).derive_time_since(now).map(|d| (format!("{:.2}", d.0), d.1)))
			} else {
				to_json::<Option<u8>>(None)
			}
		);
		//println!("{:?}", x.0);
		x.0
	}).collect::<Vec<&serde_json::Value>>();
	HttpResponse::Ok().body(tdata.handlebars.render("board_index", &btreemap!(
		"boards" => to_json(boards_json),
		"user" => to_json(user)
	)).unwrap())
}

#[get("/board/{board_id}/")]
pub async fn get_board(
	req: HttpRequest,
	tdata: web::Data<ThreadData<'_>>,
	data: web::Data<AppState>,
	web::Path(board_id): web::Path<u32>
) -> HttpResponse {
	let boards = data.boards.lock().unwrap();
	let cookie = req.cookie("auth");
	let uid = if cookie.is_some() {
		UserCookie::parse(cookie.unwrap().value())
	} else {
		UserCookie::parse("null")
	};
	if let Some(a) = Board::search(boards.iter(), board_id) {
		let data = btreemap!(
			"messages"	=> {
				let mut msg_json = to_json(&a.messages);
				let mut op = Options::empty();
				op.insert(Options::ENABLE_TABLES);

				let msg_json = msg_json.as_array_mut().unwrap().iter_mut().zip(a.messages.borrow().iter()).map(|x| {
					let parse = Parser::new_ext(&x.1.text, op);
					let o = x.0.as_object_mut().unwrap();
					o.insert(
						"time_meta".to_string(),
						to_json(DateTime::<Utc>::from(x.1.timestamp).to_rfc2822().trim_matches(|c| c=='0' || c=='+' || c==' '))
					);
					o.insert(
						"render_text".to_string(),
						{
							let mut dirty_html = String::new();
							push_html(&mut dirty_html, parse);
							to_json(dirty_html)
						}
					);
					x.0
				}).collect::<Vec<&mut serde_json::Value>>();
				to_json(msg_json)
			},
			"user" => match uid {
				Ok(a) => to_json(data.user_manager.lock().unwrap().get_user_by_id(a.id)),
				Err(_) => to_json(None::<Option<u8>>)
			},
			"board" => to_json(a)
		);
		HttpResponse::Ok().body(tdata.handlebars.render("board", &data).unwrap())
	} else {
		HttpResponse::NotFound().body("That board could not be found")
	}
}

#[post("/boards/post")]
pub async fn board_post(
	req: HttpRequest,
	form: web::Form<MessageForm>,
	data: web::Data<AppState>
) -> Result<HttpResponse, HttpResponse> {
	let cookie = req.cookie("auth");
	if cookie.is_none() {
		return Err(HttpResponse::Unauthorized().body("Unauthorized: You are not logged in"));
	}
	let cookie = match UserCookie::parse(cookie.unwrap().value()) {
		Ok(a) => a,
		Err(e) => {
			println!("{}", e);
			return Err(HttpResponse::BadRequest().body("Bad Request: Could not parse cookie data.\n Not logged in, or bad cookie."));
		}
	};
	let mut mf = data.msg_factory.lock().unwrap();
	let mut b = data.boards.lock().unwrap();
	let mut board = b.iter_mut().find(|x| (x.id == form.board_id));
	let um = data.user_manager.lock().unwrap();
	match board {
		Some(ref mut a) => {
			let msg = mf.create_message({
				let user = um.get_user_by_id(form.user_id).ok_or_else(|| {
					HttpResponse::NotFound().body("Not Found: User does not exist")
				})?;
				if user.secret == cookie.secret {
					User::from(user)
				} else {
					return Err(HttpResponse::Unauthorized().body("Unauthorized: Bad login attempt, invalid auth token"));
				}
			}, clean(&form.content));
			a.add_message(msg);
			Ok(HttpResponse::SeeOther().header("location", format!("/board/{}/", a.id)).finish())
		},
		None => Err(HttpResponse::BadRequest().body("Bad Request: Board does not exist"))
	}
}

#[get("/boards/new/")]
pub async fn new_board(
	req: HttpRequest,
	data: web::Data<AppState>,
) -> HttpResponse {
	let cookie = req.cookie("auth");
	if cookie.is_none() {
		return HttpResponse::Unauthorized().body("Unauthorized: You are not logged in");
	}

	let cookie = match UserCookie::parse(cookie.unwrap().value()) {
		Ok(a) => a,
		Err(_e) => { return HttpResponse::BadRequest().body("Bad Request: Could not parse user cookie") }
	};

	let um = data.user_manager.lock().unwrap();

	if let Some(u) = um.get_user_by_id(cookie.id) {
		if cookie.secret != u.secret {
			return HttpResponse::BadRequest().body("Bad Request: Could not authenticate user");
		}
	} else {
		return HttpResponse::BadRequest().body("Bad Request: User does not exist");
	}

	HttpResponse::Ok().body(include_str!("../../static/new_board.html"))
}

#[post("/boards/new")]
pub async fn new_board_result(
	req: HttpRequest,
	data: web::Data<AppState>,
	form: web::Form<BoardForm>
) -> impl Responder {
	let cookie = req.cookie("auth");
	if cookie.is_none() {
		return HttpResponse::Unauthorized().body("Unauthorized: You are not logged in");
	}

	let cookie = match UserCookie::parse(cookie.unwrap().value()) {
		Ok(a) => a,
		Err(_e) => { return HttpResponse::BadRequest().body("Bad Request: Could not parse user cookie") }
	};

	let um = data.user_manager.lock().unwrap();

	if let Some(u) = um.get_user_by_id(cookie.id) {
		if cookie.secret != u.secret {
			return HttpResponse::BadRequest().body("Bad Request: Could not authenticate user");
		}
	} else {
		return HttpResponse::BadRequest().body("Bad Request: User does not exist");
	}

	let mut bf = data.board_factory.lock().unwrap();
	let mut path = PathBuf::from(
		format!(
			"{}/board.json", bf.next_id
		)
	);
	path.push(&form.name);
	if let Ok(_a) = Board::open(path) {
		return HttpResponse::Conflict().body("Conflict: Board already exists");
	}
	match bf.create_board(&form.name, &form.description, cookie.id) {
		Ok(a) => {
			if form.name.contains(' ') {
				return HttpResponse::BadRequest().body("Bad Request: Name cannot contain spaces");
			}
			println!("new board");
			data.boards.lock().unwrap().push(a.clone());
			HttpResponse::SeeOther().header("location", format!("/board/{}/", a.id)).finish()
		},
		Err(e) => HttpResponse::InternalServerError().body(format!("Internal Server Error: {}", e))
	}
}
